package org.gcube.application.cms.concessioni.plugins;

import lombok.extern.slf4j.Slf4j;
import org.bson.BsonDocument;
import org.bson.BsonString;
import org.bson.BsonValue;
import org.bson.Document;
import org.gcube.application.cms.plugins.faults.EventException;
import org.gcube.application.cms.plugins.reports.EventExecutionReport;
import org.gcube.application.cms.plugins.requests.EventExecutionRequest;
import org.gcube.application.cms.serialization.Serialization;
import org.gcube.application.cms.custom.gna.concessioni.model.ProfiledConcessione;
import org.gcube.application.cms.plugins.LifecycleManager;
import org.gcube.application.cms.plugins.faults.InitializationException;
import org.gcube.application.cms.plugins.faults.ShutDownException;
import org.gcube.application.cms.plugins.faults.StepException;
import org.gcube.application.cms.plugins.model.PluginDescriptor;
import org.gcube.application.cms.plugins.reports.StepExecutionReport;
import org.gcube.application.cms.plugins.reports.InitializationReport;
import org.gcube.application.cms.plugins.requests.StepExecutionRequest;
import org.gcube.application.geoportal.common.model.document.*;
import org.gcube.application.geoportal.common.model.document.access.Access;
import org.gcube.application.geoportal.common.model.document.access.AccessPolicy;
import org.gcube.application.geoportal.common.model.document.filesets.RegisteredFileSet;
import org.gcube.application.geoportal.common.model.document.lifecycle.LifecycleInformation;
import org.gcube.application.geoportal.common.model.document.lifecycle.TriggeredEvents;
import org.gcube.application.geoportal.common.model.legacy.report.ConstraintCheck;
import org.gcube.application.geoportal.common.model.rest.Configuration;
import org.gcube.application.geoportal.common.model.rest.ConfigurationException;

@Slf4j
public class ConcessioniLifeCycleManager implements LifecycleManager {

    private static final PluginDescriptor DESCRIPTOR=new PluginDescriptor("GNA-CONCESSIONI-LC", PluginDescriptor.BaseTypes.LIFECYCLE_MANAGER);
    static {
        DESCRIPTOR.setDescription("GNA Concessioni. This plugin supports custom lifecycle management for the GNA Concessioni UseCase.");
        DESCRIPTOR.setVersion(new ComparableVersion("1.0.0"));
    }


    @Override
    public InitializationReport initInContext() throws InitializationException {
        return null;
    }

    @Override
    public InitializationReport init() throws InitializationException {
        return null;
    }

    @Override
    public void shutdown() throws ShutDownException {

    }

    @Override
    public StepExecutionReport performStep(StepExecutionRequest request) throws StepException {
        log.info("Serving Request {}",request);
        StepExecutionReport report=new StepExecutionReport();
        LifecycleInformation info=new LifecycleInformation();
        info.setPhase(request.getStep());
        info.setLastOperationStatus(LifecycleInformation.Status.OK);
        report.setRequest(request);
        report.setResult(request.getDocument());

        try {
            switch (request.getStep()) {

                case StepExecutionRequest.Steps.ON_MATERIALIZE_DOCUMENT: {
                    // SDI Create Layers
                    break;
                }
                case StepExecutionRequest.Steps.ON_DEMATERIALIZE_DOCUMENT: {
                    // SDI Remove Layers
                    break;
                }
                case StepExecutionRequest.Steps.ON_DEINDEX_DOCUMENT: {
                    // SDI Remove from centroids
                    break;
                }
                case StepExecutionRequest.Steps.ON_INDEX_DOCUMENT: {
                    // SDI Add to centroids
                    break;
                }

                case "SUBMIT-FOR-REVIEW" :{
                    //Checks
                    //Materialize layers
                    //Index-confidential
                    //Notifications
                    break;
                }

                case "REJECT-DRAFT":{
                    // Notification
                    // Set
                    break;
                }

                case "APPROVE DRAFT":{
                    break;
                }

                default:
                    throw new StepException("Invalid Step " + request.getStep());
            }
        }catch (StepException e){
            throw e;
        }catch (Throwable t){
            log.error("Unable to perform step "+request.getStep(),t);
            report.getResult().getLifecycleInformation().setLastOperationStatus(LifecycleInformation.Status.ERROR);
            report.getResult().getLifecycleInformation().addErrorMessage("Unable to execute Step "+request.getStep()+". Error was "+t.getMessage());
        }
        return report;
    }


    @Override
    public EventExecutionReport onEvent(EventExecutionRequest request) throws EventException {
        log.info("Executing Event {}",request);
        EventExecutionReport report=new EventExecutionReport();
        report.setRequest(request);
        report.setResult(request.getDocument());
        TriggeredEvents info=new TriggeredEvents();
        try {
            switch(request.getEvent()){
                case EventExecutionRequest.Events.ON_INIT_DOCUMENT:
                    // Set Defaults as for on update
                case EventExecutionRequest.Events.ON_UPDATE_DOCUMENT: {
                    log.debug("Setting default values..");
                    report.setResult(setDefaults(request.getDocument()));
                    break;
                }
                case EventExecutionRequest.Events.ON_DELETE_DOCUMENT: {
                    //DELETE ALL
                    break;
                }
                default: throw new EventException("Unexpected Event "+request.getEvent());
            }
        }catch (EventException e){
            throw e;
        }catch (Throwable t){
            log.error("Unable to execute on event "+request.getEvent(),t);
            info.setLastOperationStatus(LifecycleInformation.Status.ERROR);
            info.addErrorMessage("Unable to execute on event "+request.getEvent()+". Error was "+t.getMessage());
        }finally{
            report.getResult().getLifecycleInformation().addEventReport(info);
        }
        return report;
    }

    @Override
    public Configuration getCurrentConfiguration() throws ConfigurationException {
        return null;
    }

    @Override
    public PluginDescriptor getDescriptor() {
        return DESCRIPTOR;
    }


    // Materialization
    private static final void materialize(){

    }


    // STATIC ROUTINES

    private static final ProfiledDocument setDefaults(ProfiledDocument document){
        ProfiledConcessione c=Serialization.convert(document,ProfiledConcessione.class);



        Document doc=c.getTheDocument();
        doc.putIfAbsent(ProfiledConcessione.SOGGETTO,new String[]{"Research Excavation","Archaeology"});

        doc.putIfAbsent(ProfiledConcessione.DSCRIZIONE_CONTENUTO,"Relazione di fine scavo e relativo abstract; selezione di immagini rappresentative;"
                + " posizionamento topografico dell'area indagata, pianta di fine scavo.");

        // Super Section
        c.getInfo().getAccess().setLicense(
                ConstraintCheck.defaultFor(c.getInfo().getAccess().getLicense(), "CC0-1.0").evaluate());


        //RELAZIONE
        // TODO NB provare se object gia' presente sia usando Document che sub Object
        doc.putIfAbsent(ProfiledConcessione.RELAZIONE_SCAVO,new Document());
        Document rel=Serialization.convert(doc.get(ProfiledConcessione.RELAZIONE_SCAVO), Document.class);

        rel.putIfAbsent(ProfiledConcessione.Sections.TITOLO,doc.getString(ProfiledConcessione.NOME)+" relazione di scavo");
        rel.putIfAbsent(ProfiledConcessione.SOGGETTO,doc.get(ProfiledConcessione.SOGGETTO));
        rel.putIfAbsent(RegisteredFileSet.CREATION_INFO,c.getInfo().getCreationInfo());
        rel.putIfAbsent(RegisteredFileSet.ACCESS,c.getInfo().getAccess());
        Access relAccess=Serialization.convert(rel.get(RegisteredFileSet.ACCESS),Access.class);
        relAccess.setLicense(ConstraintCheck.defaultFor(relAccess.getLicense(),"CC-BY-4.0").evaluate());
        relAccess.setPolicy(ConstraintCheck.defaultFor(relAccess.getPolicy(), AccessPolicy.OPEN).evaluate());

        //ABSTRACT Relazione
        doc.putIfAbsent(ProfiledConcessione.ABSTRACT_RELAZIONE,new Document());
        Document abs=Serialization.convert(doc.get(ProfiledConcessione.ABSTRACT_RELAZIONE),Document.class);
        abs.putIfAbsent(ProfiledConcessione.Sections.TITOLO,doc.getString(ProfiledConcessione.NOME)+" abstract relazione di scavo");
        abs.putIfAbsent(RegisteredFileSet.CREATION_INFO,c.getInfo().getCreationInfo());
        abs.putIfAbsent(RegisteredFileSet.ACCESS,c.getInfo().getAccess());
        Access absAccess=Serialization.convert(abs.get(RegisteredFileSet.ACCESS),Access.class);
        absAccess.setLicense(ConstraintCheck.defaultFor(absAccess.getLicense(),"CC-BY-4.0").evaluate());
        absAccess.setPolicy(ConstraintCheck.defaultFor(absAccess.getPolicy(), AccessPolicy.OPEN).evaluate());

        //TODO complete setDefaults
        //IMMAGINI RAPPRESENTATIVE


        if(doc.containsKey(ProfiledConcessione.IMMAGINI_RAPPRESENTATIVE)) {
            for (BsonValue bsonValue : doc.toBsonDocument(null, null).
                    getArray(ProfiledConcessione.IMMAGINI_RAPPRESENTATIVE)) {

                BsonDocument imgDocument = bsonValue.asDocument();
                imgDocument.putIfAbsent(ProfiledConcessione.SOGGETTO,new BsonString(doc.getString(ProfiledConcessione.SOGGETTO)));
//                imgDocument.putIfAbsent(RegisteredFileSet.CREATION_INFO,this.getInfo().getCreationInfo());
//                imgDocument.putIfAbsent(RegisteredFileSet.CREATION_INFO,new BsonDocument(new Document()));
//                imgDocument.putIfAbsent(RegisteredFileSet.ACCESS,this.getInfo().getAccess());
//                Access absAccess=rel.get(RegisteredFileSet.ACCESS,Access.class);
//                absAccess.setLicense(ConstraintCheck.defaultFor(absAccess.getLicense(),"CC-BY-4.0").evaluate());
//                absAccess.setPolicy(ConstraintCheck.defaultFor(absAccess.getPolicy(), AccessPolicy.OPEN).evaluate());
            }
        }


        return c;
    }


}
